%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Create main datasets for the study by first cleaning them with Python,
% and then estimating yield curves for Norway and Sweden with Matlab.
% Created by Tuomas Tomunen
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
close all;
clear all;

%%
% Clean data with Python
system('python create_datasets_helper.py');

%%
% Estimate Norwegian yield curve with Svensson method
no_data  = readtable('norway_TEMP.csv');
delete 'norway_TEMP.csv'

no_yield = table2array(no_data(:, 2:end));
no_yieldmat = zeros(length(no_yield), 5);
coupon_rate = [0; 0; 0; 0; 0; 0; 0];

% Settle date and maturity dates are arbitrary - only difference matters
settle_date   = repmat(datenum('31-Dec-1989'), size(no_yield, 2), 1);
maturity_date = [datenum('31-Mar-1990'); % 3-month
                datenum('30-Jun-1990');  % 6-month
                datenum('30-Sep-1990');  % 9-month
                datenum('31-Dec-1990');  % 1-year
                datenum('31-Dec-1992');  % 3-year
                datenum('31-Dec-1994');  % 5-year
                datenum('31-Dec-1999')]; % 10-year
            
out_dates = [datenum('31-Dec-1990');  % 1-year
             datenum('31-Dec-1991');  % 2-year
             datenum('31-Dec-1992');  % 3-year
             datenum('31-Dec-1993');  % 4-year
             datenum('31-Dec-1994')]; % 5-year


for t = 1 : length(no_yield)
    % Convert yields at particular date to prices.
    % Norway uses annual compounding frequency.
    price = bndprice(no_yield(t, :) / 100, coupon_rate, ...
                     settle_date, maturity_date, ...
                     'CompoundingFrequency', 1);
    % Fit yield curve with Svensson function from Financial Instruments
    % toolbox
    curve = IRFunctionCurve.fitSvensson('Zero', settle_date(1), ...
        [settle_date maturity_date price coupon_rate], 'Compounding', -1);
    no_yieldmat(t, :) = curve.getZeroRates(out_dates)';
end

% Supplement Norway data with Wright (2011) data between 1998 and 2002.
no_wright = xlsread('raw_data/ZCdata.xls', 'Norway', 'A3:V62'); % Data from 1/1998 to 12/2002 and maturities from 3 to 60 months
no_wrighttable = array2table(no_wright(:, [6, 10, 14, 18, 22]), 'VariableNames',{'NOK1Y','NOK2Y','NOK3Y', 'NOK4Y', 'NOK5Y'});

date = eomdate(datetime([no_wright(:,1), no_wright(:,2), ones(length(no_wright), 1)]));
no_wrighttable = [table(date, 'VariableNames', {'date'}), no_wrighttable];

no_yieldtable = array2table(no_yieldmat*100, 'VariableNames',{'NOK1Y','NOK2Y','NOK3Y', 'NOK4Y', 'NOK5Y'});
no_yieldtable = [table(no_data{:, 'date'}, 'VariableNames', {'date'}) no_yieldtable];

no_yieldtable = [no_wrighttable; no_yieldtable];

writetable(no_yieldtable, 'NOK.csv')

%%
% Estimate Swedish yield curve with Svensson method
swe_data  = readtable('sweden_TEMP.csv');
delete 'sweden_TEMP.csv'

% Estimation must be done separately for pre and post 6/2009 periods,
% because 1-year yield is missing afterwards
cutoff = find(table2array(swe_data(:, {'date'}))=='2009-05-31');

% Pre
swe_yield = table2array(swe_data(1:cutoff, 2:end));
swe_yieldmat = zeros(size(swe_data, 1), 5);
coupon_rate = zeros(size(swe_yield, 2), 1);

% Settle date and maturity date are arbitrary - only difference matters
settle_date   = repmat(datenum('31-Dec-1989'), size(swe_yield, 2), 1);
maturity_date = [datenum('31-Jan-1990');  % 1-month
                 datenum('31-Mar-1990');  % 3-month
                 datenum('30-Jun-1990');  % 6-month
                 datenum('31-Dec-1990');  % 1-year
                 datenum('31-Dec-1991');  % 2-year
                 datenum('31-Dec-1994');  % 5-year
                 datenum('31-Dec-1996');  % 7-year                 
                 datenum('31-Dec-1999')]; % 10-year
            
out_dates = [datenum('31-Dec-1990');  % 1-year
             datenum('31-Dec-1991');  % 2-year
             datenum('31-Dec-1992');  % 3-year
             datenum('31-Dec-1993');  % 4-year
             datenum('31-Dec-1994')]; % 5-year

for i = 1 : length(swe_yield)
    % Convert yields at particular date to prices
    % Sweden uses annual compounding frequency.
    price = bndprice(swe_yield(i, :) / 100, coupon_rate, ...
                     settle_date, maturity_date, ...
                     'CompoundingFrequency', 1);
                    
    % Fit yield curve with Svensson function from Financial Instruments
    % toolbox
    curve = IRFunctionCurve.fitSvensson('Zero', settle_date(1), ...
        [settle_date maturity_date price coupon_rate], 'Compounding', -1);
    swe_yieldmat(i, :) = curve.getZeroRates(out_dates)';
end

% Post
swe_yield = table2array(swe_data(cutoff+1:end, 2:end));
swe_yield = [swe_yield(:, 1:3) swe_yield(:, 5:end)];
coupon_rate = zeros(size(swe_yield, 2), 1);

% Settle date and maturity date are arbitrary - only difference matters
settle_date   = repmat(datenum('31-Dec-1989'), size(swe_yield, 2), 1);
maturity_date = [datenum('31-Jan-1990');  % 1-month
                 datenum('31-Mar-1990');  % 3-month
                 datenum('30-Jun-1990');  % 6-month
                 datenum('31-Dec-1991');  % 2-year
                 datenum('31-Dec-1994');  % 5-year
                 datenum('31-Dec-1996');  % 7-year                 
                 datenum('31-Dec-1999')]; % 10-year
            
out_dates = [datenum('31-Dec-1990');  % 1-year
             datenum('31-Dec-1991');  % 2-year
             datenum('31-Dec-1992');  % 3-year
             datenum('31-Dec-1993');  % 4-year
             datenum('31-Dec-1994')]; % 5-year

for i = 1 : length(swe_yield)
    price = bndprice(swe_yield(i, :) / 100, coupon_rate, ...
                     settle_date, maturity_date, ...
                     'CompoundingFrequency', 1);
    curve = IRFunctionCurve.fitSvensson('Zero', settle_date(1), ...
        [settle_date maturity_date price coupon_rate], 'Compounding', -1);
    swe_yieldmat(cutoff+i, :) = curve.getZeroRates(out_dates)';
end

swe_yieldmat = swe_yieldmat*100;

swe_yieldtable = array2table(swe_yieldmat, 'VariableNames',{'SEK1Y','SEK2Y','SEK3Y', 'SEK4Y', 'SEK5Y'});
swe_yieldtable = [table(swe_data{:, 'date'}, 'VariableNames', {'date'}) swe_yieldtable];

writetable(swe_yieldtable, 'SEK.csv')
