import os

import numpy as np
import pandas as pd
import pandas_datareader.data as web
from pandas.tseries.offsets import MonthEnd

import create_us

END_DATE = pd.to_datetime('12-31-2016')  # Last period for all series


def main():
    """Run all country-specific data cleaning scripts in this module."""
    norway()
    sweden()
    au_jp_nz()
    us()
    uk()
    eu()
    switzerland()
    canada()
    exchange_rates()

    pass


def norway():
    """
    Construct end-of-month yield series from daily series using the last available observations in each month.
    To get correct maturities, the output must be used to fit a yield curve with Matlab.
    """
    START_DATE = pd.to_datetime('01-08-2003')  # Before 2003, data is not available for all maturities.

    df = pd.read_csv('raw_data/renter_dag.csv')
    df = df.replace(['ND', 'NC'], np.nan)

    df = df[['DATES',
             'STATSVKL.3M.EFF',
             'STATSVKL.6M.EFF',
             'STATSVKL.9M.EFF',
             'STATSVKL.12M.EFF',
             'STATSOBL.3Y.EFF',
             'STATSOBL.5Y.EFF',
             'STATSOBL.10Y.EFF']]

    df.columns = ['date',
                  'NOK3M',
                  'NOK6M',
                  'NOK9M',
                  'NOK1Y',
                  'NOK3Y',
                  'NOK5Y',
                  'NOK10Y']

    df['date'] = pd.to_datetime(df['date'])
    df = df[(df['date'] >= START_DATE) & (df['date'] <= END_DATE)]

    df = df.fillna(method='pad', limit=30)

    # Change last observation (12/30/2016) to month end (12/31/2016). Assumes last obs is in the last row.
    df.loc[df.index.max(), 'date'] = df['date'].max() + MonthEnd(0)
    df = df.set_index('date').asfreq('M', method='pad', how='last')

    df.to_csv('norway_TEMP.csv')

    pass


def sweden():
    """
    Construct end-of-month yield series from daily series using the last available observations in each month.
    To get correct maturities, the output must be used to fit a yield curve with Matlab.
    """
    START_DATE = pd.to_datetime('01-01-1987')  # Before 1987, data is not available for all maturities.

    files = ['ADVANCED_RESULT_28_09_2017_4_00_PM_80_2',
             'ADVANCED_RESULT_28_09_2017_4_01_PM_90_1',
             'ADVANCED_RESULT_28_09_2017_4_02_PM_90_2',
             'ADVANCED_RESULT_28_09_2017_4_03_PM_00_1',
             'ADVANCED_RESULT_28_09_2017_4_04_PM_00_2',
             'ADVANCED_RESULT_28_09_2017_4_05_PM_10_1']

    df = pd.DataFrame()
    for i in files:
        df2 = pd.read_csv('raw_data/{}.csv'.format(i), sep=';')
        df2 = df2.replace('n/a', np.nan)
        df2 = df2.pivot(index='Date', columns='Series', values='Value').reset_index()
        df = df.append(df2)

    df['Date'] = pd.to_datetime(df['Date'], dayfirst=True)
    df = df[['Date', 'SE 1M', 'SE 3M', 'SE 6M', 'SE 12M', 'SE GVB 2Y',
             'SE GVB 5Y', 'SE GVB 7Y', 'SE GVB 10Y']]

    df.columns = ['date',
                  'SEK1M',
                  'SEK3M',
                  'SEK6M',
                  'SEK1Y',
                  'SEK2Y',
                  'SEK5Y',
                  'SEK7Y',
                  'SEK10Y']

    df = df.sort_values('date')

    df = df[(df['date'] >= START_DATE) & (df['date'] <= END_DATE)]

    # There was a bank strike from 1/25/1990 to 2/19/1990. Jan 1990 observation is hence prior to strike.
    df = df.fillna(method='pad', limit=30).reset_index(drop=True)

    df.loc[df.index.max(), 'date'] = df['date'].max() + MonthEnd(0)
    df = df.set_index('date').asfreq('M', method='pad', how='last')

    df.index = df.index + MonthEnd(0)

    df.loc[df.index > pd.to_datetime('06-01-2009'), 'SEK1Y'] = np.nan  # Last frequent observation in 6/4/2009.

    df.to_csv('sweden_TEMP.csv')

    pass


def au_jp_nz():
    """
    Construct end-of-month yield series for Australia, Japan and New Zealand by first constructing monthly
    series from from daily series using the last available observations in each month using data from Datastream,
    and then backfilling it with Wright (2011) data.

    A switch to Datastream data is made immediately when it becomes available for all maturities
    (NZ is missing 4-year data).
    """
    COUNTRY_DATA = {'Australia':
                    {'currency': 'AUD',
                     'start_date': pd.to_datetime('03-14-2005')},
                    'Japan':
                    {'currency': 'JPY',
                     'start_date': pd.to_datetime('09-20-2000')},
                    'New Zealand':
                    {'currency': 'NZD',
                     'start_date': pd.to_datetime('06-13-1996')}}

    for i, data in COUNTRY_DATA.items():
        df = pd.read_excel('raw_data/jp_au_nz_yields.xlsx', sheet_name=i, header=1)

        df.columns = ['date'] + ['{}{}Y'.format(data['currency'], n) for n in range(1, 6)]

        df = df[(df['date'] >= data['start_date']) & (df['date'] <= END_DATE)]
        df = df.fillna(method='pad', limit=30)

        # Change last observation to month end, assuming that last observation is in the last row
        df.loc[df.index.max(), 'date'] = df['date'].max() + MonthEnd(0)

        df = df.set_index('date').asfreq('M', method='pad', how='last')

        df = np.log(1 + df / 200) * 200  # All rates are semiannually compounded.

        wright_df = pd.read_excel('raw_data/ZCdata.xls', sheet_name=i, header=1).reset_index()

        wright_df['date'] = pd.to_datetime(wright_df['level_0']*10000
                                           + wright_df['level_1']*100
                                           + 1, format='%Y%m%d') + MonthEnd(0)

        wright_df = wright_df[['date', 12, 24, 36, 48, 60]]

        wright_df.columns = ['date'] + ['{}{}Y'.format(data['currency'], n) for n in range(1, 6)]

        wright_df = wright_df[wright_df['date'] < df.index.min()]
        wright_df = wright_df.set_index('date').append(df)

        wright_df.to_csv('{}.csv'.format(data['currency']))

    pass


def us():
    create_us.create_us()
    pass


def uk():
    START_DATE = pd.to_datetime('06-30-1952')

    df = pd.read_excel('raw_data/GBP_RATES.xlsx',
                       sheet_name='4. spot curve', header=None, skiprows=5, usecols=10,
                       names=['date', 'GBP0_5Y', 'GBP1Y', 'GBP1_5Y', 'GBP2Y', 'GBP2_5Y', 'GBP3Y',
                              'GBP3_5Y', 'GBP4Y', 'GBP4_5Y', 'GBP5Y'])

    df = df[(df['date'] >= START_DATE) & (df['date'] <= END_DATE)]

    df[['date', 'GBP1Y', 'GBP2Y', 'GBP3Y', 'GBP4Y', 'GBP5Y']].to_csv('GBP.csv', index=False)

    pass


def eu():
    START_DATE = pd.to_datetime('01-01-1973')
    df = pd.read_excel('raw_data/EUR_RATES.xlsx', sheet_name='Sheet1', usecols=6)

    df['date'] = pd.to_datetime(df['Year']*10000+df['Month']*100+1, format='%Y%m%d') + MonthEnd(0)
    df = df[(df['date'] >= START_DATE) & (df['date'] <= END_DATE)]

    df[['date', 'EUR1Y', 'EUR2Y', 'EUR3Y', 'EUR4Y', 'EUR5Y']].to_csv('EUR.csv', index=False)

    pass


def switzerland():
    START_DATE = pd.to_datetime('01-01-1988')
    df = pd.read_excel('raw_data/CHF_RATES.xlsx', sheet_name='raw rates', usecols=6)
    df = df.rename(columns={'CHF12': 'CHF1Y', 'CHF24': 'CHF2Y', 'CHF36': 'CHF3Y', 'CHF48': 'CHF4Y', 'CHF60': 'CHF5Y'})

    df['date'] = pd.to_datetime(df['Year']*10000+df['Month']*100+1, format='%Y%m%d') + MonthEnd(0)
    df = df[(df['date'] >= START_DATE) & (df['date'] <= END_DATE)]

    df[['date', 'CHF1Y', 'CHF2Y', 'CHF3Y', 'CHF4Y', 'CHF5Y']].to_csv('CHF.csv', index=False)

    pass


def canada():
    START_DATE = pd.to_datetime('01-01-1986')
    df = pd.read_excel('raw_data/Canada_updated.xlsx', sheet_name='Sheet1', usecols=6)

    df['date'] = pd.to_datetime(df['Year']*10000+df['Month']*100+1, format='%Y%m%d') + MonthEnd(0)
    df = df[(df['date'] >= START_DATE) & (df['date'] <= END_DATE)]

    df[['CAD1Y', 'CAD2Y', 'CAD3Y', 'CAD4Y', 'CAD5Y']] *= 100

    df[['date', 'CAD1Y', 'CAD2Y', 'CAD3Y', 'CAD4Y', 'CAD5Y']].to_csv('CAD.csv', index=False)

    pass


def exchange_rates():
    START_DATE = pd.to_datetime('01-01-1971')
    END_DATE = pd.to_datetime('06-30-2017')

    # SERIES contains all exchange rate series. Key is series name in FRED, first element of value is column name ion
    # output. Second element is True if series is expressed in USD/[Other Curr] instead of [Other Curr]/USD.
    # In this case, invert series.
    SERIES = {'DEXCAUS': ['CADUSD', False],
              'DEXUSEU': ['EURUSD', True],
              'DEXJPUS': ['JPYUSD', False],
              'DEXNOUS': ['NOKUSD', False],
              'DEXSDUS': ['SEKUSD', False],
              'DEXSZUS': ['CHFUSD', False],
              'DEXUSAL': ['AUDUSD', True],
              'DEXUSUK': ['GBPUSD', True],
              'DEXUSNZ': ['NZDUSD', True]}

    df = pd.DataFrame()
    for name, i in SERIES.items():
        df2 = web.DataReader(name, 'fred', start=START_DATE, end=END_DATE).reset_index()
        df2 = df2.rename(columns={'DATE': 'date', name: i[0]})
        if i[1]:
            df2[i[0]] = 1 / df2[i[0]]
        if df.empty:
            df = df2
        else:
            df = df.merge(df2, on='date', how='outer', validate='1:1')

    df3 = pd.read_csv('raw_data/bbk_paket1.csv', header=None, engine='python',
                      names=['date', 'USDEUR', 'empty1', 'DEMUSD', 'empty2', 'DEMEUR', 'empty3'],
                      skiprows=5, skipfooter=11, na_values=['.'],
                      parse_dates=['date'])[['date', 'USDEUR', 'DEMUSD', 'DEMEUR']]

    assert df3['DEMEUR'].notnull().sum() == 1
    EURDEM = 1 / df3['DEMEUR'].max()
    df3['EURUSD'] = 1 / df3['USDEUR']
    df3.loc[df3['date'] <= '1/1/1999', 'EURUSD'] = EURDEM * df3['DEMUSD']
    df3 = df3[(df3['date'] >= START_DATE) & (df3['date'] <= END_DATE)]
    df = df.merge(df3[['date', 'EURUSD']], on='date', how='left', validate='1:1', suffixes=['', '_buba'])

    assert df[['EURUSD', 'EURUSD_buba']].corr().iloc[0, 1] >= 0.9996
    df.loc[df['date'] <= '1/1/1999', 'EURUSD'] = df['EURUSD_buba']
    df = df.drop(columns=['EURUSD_buba'])

    df = df.fillna(method='pad', limit=30).reset_index(drop=True)

    df.loc[df.index.max(), 'date'] = df['date'].max() + MonthEnd(0)
    df = df.set_index('date').asfreq('M', method='pad', how='last')

    df.to_csv('FX_to_USD.csv')

    pass


main()
