import numpy as np
import pandas as pd
import statsmodels.formula.api as sm
from scipy.stats import chi2

from modules import load_data
from modules.utils.S_matrix import S_matrix


def run(df, y_variables, x_variables, include_constant_1stage=True):
    """
    Run two-stage Cochrane-Piazzesi regressions given excess bond returns and forward spreads.

    Parameters
    ----------
    df: pandas.DataFrame
        A dataframe that contains time series for all excess bond returns and forward spreads (for one country).
    y_variables: list
        A list of df column names that contain excess bond returns used in the analysis.
    x_variables: list
        A list of df column names that contain forward spreads used in the analysis.
    include_constant_1stage: boolean, default True
        If False, exclude constant from first-stage regression.

    Output
    ------
    A dictionary with following items:

    gammas: pandas.DataFrame
        First-stage parameter estimates and their GMM standard errors.
    bs: pandas.DataFrame
        Second-stage parameter estimates and their GMM standard errors.
    rsquared_adj: numpy.float64
        First-stage adjusted R-squared.
    wald_test: dictionary
        Results of a Wald test that tests if all first-stage parameters  (excluding constant) are jointly zero.
    VCV: numpy.matrixlib.defmatrix.matrix
        Variance-covariance matrix of gammas and bs.
    1stage:  pandas.DataFrame
        First-stage residuals and fitted values (i.e. Cochrane-Piazzesi factor).
    """
    df = df.copy()

    df['y_bar'] = df[y_variables].mean(axis=1)

    output = {}

    # 1. First stage
    if include_constant_1stage:
        formula = 'y_bar ~ {}'.format(' + '.join(x_variables))  # e.g. 'y_bar ~ fs2_bar + fs3_bar'
    else:
        formula = 'y_bar ~ -1 + {}'.format(' + '.join(x_variables))  # e.g. 'y_bar ~ fs2_bar + fs3_bar'
    ols = sm.ols(formula, data=df).fit(cov_type='HAC', use_t=False, cov_kwds={'maxlags': 18})

    output['gammas'] = ols.params.to_frame('params')
    output['rsquared_adj'] = ols.rsquared_adj

    stderrs = ols.bse
    res_1stage = np.matrix(ols.resid.values).T
    df['cp_factor'] = ols.fittedvalues
    df['resid_1stage'] = ols.resid

    output['1stage'] = df[['date', 'cp_factor', 'resid_1stage']].reset_index(drop=True)

    # Run Wald test (for diagnostic purposes)
    # formula = ', '.join('({} = 0)'.format(i) for i in x_variables)  # e.g. '(fs2_bar = 0), (fs3_bar = 0)'
    # wald_stat = np.asscalar(ols.wald_test(formula, use_f=False).statistic)

    # 2. Second stage
    b = []
    res_2stage = []
    for n in y_variables:
        ols = sm.ols('{} ~ -1 + cp_factor'.format(n), data=df).fit()
        b += [ols.params['cp_factor']]
        res_2stage += [list(ols.resid.values)]
    b = np.array([b]).T
    output['bs'] = pd.DataFrame(b, columns=['params'], index=y_variables)
    res_2stage = np.array(res_2stage).T

    # 3. Calculate GMM standard errors. See page 15 of Cochrane-Piazzesi (2005) online appendix for formulas.
    if include_constant_1stage:
        f_T = np.hstack([np.ones([df.shape[0], 1]), df[x_variables].values])
    else:
        f_T = df[x_variables].values
    y_T = df[y_variables].values
    gamma = np.matrix(output['gammas']['params'].values).T

    T = f_T.shape[0]
    N_y = y_T.shape[1]
    N_x = f_T.shape[1]

    EffT = (f_T.T @ f_T) / T

    d_mat = np.hstack([-EffT, np.zeros([N_x, N_y])])
    d_mat = np.vstack([d_mat, np.hstack([y_T.T @ f_T / T - 2 * b @ gamma.T @ EffT,
                                         -np.asscalar(gamma.T @ EffT @ gamma) * np.eye(N_y)])])

    u = np.multiply(np.tile(res_1stage, [1, N_x]), f_T)
    u = np.hstack([u, np.multiply(np.tile(np.matrix(df['cp_factor'].values).T, [1, N_y]), res_2stage)])

    R = np.hstack([np.zeros([N_x - 1, 1]), np.eye(N_x - 1)])  # R-matrix for Wald test

    # Diagnostic test: GMM standard errors for gammas should be standard Newey-West ones (if no demeaning for S_matrix).
    # Compare them and resulting Wald-statistic to statmodels' output.
    # S_temp = S_matrix(u, 18, newey_west=True, demean=False)
    # VCV_temp = np.linalg.inv(d_mat) @ S_temp @ np.linalg.inv(d_mat).T / T
    # stderrs_temp = np.sqrt(np.diag(VCV_temp))
    # assert (np.round(stderrs_temp[0:N_x], 6) == np.round(stderrs.values, 6)).all()
    # assert np.round((R @ gamma).T @ np.linalg.inv((R @ VCV_temp[:N_x, :N_x] @ R.T)) @ (R @ gamma), 5) \
    #     == np.round(wald_stat, 5)

    S = S_matrix(u, 18, newey_west=True, demean=True)
    VCV = np.linalg.inv(d_mat) @ S @ np.linalg.inv(d_mat).T / T
    stderrs = np.sqrt(np.diag(VCV))

    output['gammas']['se'] = stderrs[:N_x]
    output['bs']['se'] = stderrs[N_x:]
    output['VCV'] = VCV

    chi2_stat = np.asscalar((R @ gamma).T @ np.linalg.inv((R @ VCV[:N_x, :N_x] @ R.T)) @ (R @ gamma))
    output['wald_test'] = {'value': chi2_stat}
    output['wald_test']['p_value'] = 1 - chi2.cdf(chi2_stat, N_x - 1)

    return output


def add_cp_factor_and_resid(df, y_variables, x_variables):
    """
    Run two-stage Cochrane-Piazzesi regressions and add CP-factor and first-stage residual as columns to df
    (named 'cp_factor' and 'resid_1stage').

    Parameters
    ----------
    df: pandas.DataFrame
        A dataframe that contains time series for all excess bond returns and forward spreads.
    y_variables: list
        A list of df column names that contain excess bond returns used in the analysis.
    x_variables: list
        A list of df column names that contain forward spreads used in the analysis.
    """
    cp_result = run(df, y_variables, x_variables)
    df = df.merge(cp_result['1stage'], on='date', validate='1:1', how='left')

    return df


def calc_dif_se(y_variables, x_variables):
    """
    Calculate difference in squared prediction errors between sample mean and Cochrane-Piazzesi model for each country.

    Parameters
    ----------
    y_variables: list
        A list of df column names that contain excess bond returns used in the analysis.
    x_variables: list
        A list of df column names that contain forward spreads used in the analysis.
    """
    output = {}
    df = load_data.load()
    for i in df['country'].unique():
        df2 = df[df['country'] == i].copy()
        cp_result = run(df2, y_variables, x_variables)

        pred_model = cp_result['1stage'][['cp_factor']].values @ cp_result['bs'][['params']].values.T

        se_model = np.power((df2[y_variables].values - pred_model), 2)

        se_hm = np.power((df2[y_variables].values - df2[y_variables].mean(axis=0).values), 2)

        output[i] = pd.DataFrame(se_hm - se_model)
        output[i].columns = y_variables
        output[i].insert(0, 'date', df2['date'].values)

    return output
