import numba
import numdifftools as nd
import numpy as np
import pandas as pd
from scipy.optimize import minimize
from scipy.stats import chi2

from modules.analyses import cp2stage
from modules.utils.S_matrix import S_matrix


def run(df, y_variables, x_variables, include_se=True, twostage=True):
    """
    Run a modified Cochrane-Piazzesi model with free constant
    described in Section 1 of Online Appendix to Hodrick-Tomunen.

    Parameters
    ----------
    df: pandas.DataFrame
        A dataframe that contains time series for all excess bond returns and forward spreads (for one country).
    y_variables_cp: list
        A list of df column names that contain excess bond returns used in the analysis.
    x_variables_cp: list
        A list of df column names that contain forward spreads used in the analysis.
    include_se: boolean, default True
        If False, skip standard error estimation to improve execution time.
    twostage: boolean, default True
        If True, run two-stage GMM. Otherwise, run one-stage using identity matrix as a weighting matrix.

    Output
    ------
    A dictionary with following items:

    alphas: pandas.DataFrame
        alpha estimates and their GMM standard errors.
    deltas: pandas.DataFrame
        delta estimates and their GMM standard errors.
    gammas: pandas.DataFrame
        gamma estimates and their GMM standard errors.
    bs: pandas.DataFrame
        b estimates and their GMM standard errors.
    rsquared: numpy.float64
        R-squared corresponding to the projection of the average returns on the forward spreads.
    J_test: dictionary
        Statistic and p-value for overidentification test (only present if twostage=True).
    """
    output = {}

    Ny = len(y_variables)
    Nx = len(x_variables)
    Northogs = Ny + Nx + Ny * Nx
    Nparams = Ny + Nx + Nx + (Ny - 1)

    y_data = df[y_variables].values
    x_data = df[x_variables].values

    cp_result = cp2stage.run(df, y_variables, x_variables)
    x0 = np.concatenate((y_data.mean(axis=0),
                         x_data.mean(axis=0),
                         cp_result['gammas']['params'].values[1:],
                         np.ones(y_data.shape[1] - 1)))

    W0 = np.eye(Northogs)

    theta1 = minimize(gmm_obj, x0, args=(W0, y_data, x_data), method='L-BFGS-B').x
    g1 = g_fun(theta1, y_data, x_data)

    if twostage:
        S1 = S_matrix(g1, 18, newey_west=True, demean=True)
        W1 = np.linalg.inv(S1)

        # Second stage
        theta2 = minimize(gmm_obj, theta1, args=(W1, y_data, x_data), method='L-BFGS-B').x
        g2 = g_fun(theta2, y_data, x_data)

        # J-test
        J = np.asscalar(gmm_obj(theta2, W1, y_data, x_data))
        output['J_test'] = {'value': J}
        output['J_test']['dofs'] = Northogs - Nparams
        output['J_test']['p_value'] = 1 - chi2.cdf(J, Northogs - Nparams)
    else:
        W1 = W0
        theta2 = theta1
        g2 = g1
    # R2
    alphas = theta2[:Ny]
    deltas = theta2[Ny:Ny+Nx]
    gammas = theta2[Ny+Nx:Ny+2*Nx]
    bs = theta2[Ny+2*Nx:]
    bs2 = np.append(bs, [(len(bs) + 1) - bs.sum()])

    y = df[y_variables].mean(axis=1)
    SS_tot = ((y - y.mean()) ** 2).sum()

    y_bar_pred = ((np.matrix(bs2).T @ (np.matrix(gammas) @ (x_data - deltas).T)).T + alphas).mean(axis=1)
    res = y_bar_pred - np.matrix(y.values).T

    SS_err = np.power(res, 2).sum()

    output['rsquared'] = 1 - SS_err / SS_tot

    # Parameters
    output['alphas'] = pd.DataFrame(alphas, columns=['params'])
    output['deltas'] = pd.DataFrame(deltas, columns=['params'])
    output['gammas'] = pd.DataFrame(gammas, columns=['params'])
    output['bs'] = pd.DataFrame(bs2, columns=['params'])

    # Standard errors
    if include_se:
        S2 = S_matrix(g2, 18, newey_west=True, demean=True)
        d_mat = nd.Jacobian(g_T, order=6)(theta2, y_data, x_data)

        VCV = np.linalg.inv(d_mat.T @ np.linalg.inv(S2) @ d_mat) / y_data.shape[0]
        stderrs = np.sqrt(np.diag(VCV))

        output['alphas']['se'] = stderrs[:Ny]
        output['deltas']['se'] = stderrs[Ny:Ny+Nx]
        output['gammas']['se'] = stderrs[Ny+Nx:Ny+2*Nx]
        output['bs']['se'] = np.append(stderrs[Ny+2*Nx:], np.nan)

    return output


@numba.jit
def gmm_obj(x, W, y_data, x_data):
    g = g_fun(x, y_data, x_data).mean(axis=0)
    return y_data.shape[0] * g @ W @ g.T


@numba.jit
def g_T(x, y_data, x_data):
    return g_fun(x, y_data, x_data).mean(axis=0)


def g_fun(x, y_data, x_data):
    x = np.squeeze(np.asarray(x))
    ixend_a = y_data.shape[1]
    ixend_d = ixend_a + x_data.shape[1]
    ixend_g = ixend_d + x_data.shape[1]

    alphas = x[0:ixend_a]
    deltas = x[ixend_a:ixend_d]
    gammas = x[ixend_d:ixend_g]
    bs = x[ixend_g:]
    bs2 = np.append(bs, [(len(bs) + 1) - bs.sum()])  # Average b is 1.

    epsilon1 = (y_data - alphas)
    epsilon2 = (x_data - deltas)
    epsilon3 = np.asarray(epsilon1 - (np.matrix(bs2).T @ (np.matrix(gammas) @ epsilon2.T)).T)

    g = np.hstack((epsilon1, epsilon2))

    for col1 in range(0, x_data.shape[1]):
        for col2 in range(0, epsilon3.shape[1]):
            g = np.hstack((g, np.matrix(np.multiply(x_data[:, col1], epsilon3[:, col2])).T))

    return g
