import numdifftools as nd
import numpy as np
import statsmodels.formula.api as sm
from scipy.stats import chi2

from modules import load_data
from modules.utils.S_matrix import S_matrix


def run(df, y_variables_cp, x_variables_cp, second_stage_spec, countries, include_se=True):
    """
    Run two-stage regressions to explain currency market excess returns with CP factors
    and additional variables, taking into account that CP factors are estimated with error.

    Parameters
    ----------
    df: pandas.DataFrame
        A dataframe that contains time series for all excess bond returns and forward spreads
        (for all countries in 'countries').
    y_variables_cp: list
        A list of df column names that contain excess bond returns used in the analysis.
    x_variables_cp: list
        A list of df column names that contain forward spreads used in the analysis.
    second_stage_spec: {'t4', 't5'}
        Specification for second-stage regression (see Table 4 and Table 5 descriptions).
    countries: list
        Defines the currency pair for which the analysis is run. Format: [target_currency (j), base_currency (i)].
        Example: ['GBP', 'USD'].
    include_se: boolean, default True
        If False, skip standard error estimation and Wald tests to improve execution time.

    Output
    ------
    A dictionary with following items:

    params: pandas.DataFrame
        Second-stage parameter estimates and their GMM standard errors.
    rsquared_adj: numpy.float64
        Second-stage adjusted R-squared.
    wald_test_i: dictionary
        Results of a Wald test that tests if base country second-stage parameters are
        jointly zero.
    wald_test_j: dictionary
        Results of a Wald test that tests if target country second-stage parameters are
        jointly zero.
    wald_test_all: dictionary
        Results of a Wald test that tests if all second-stage parameters (excluding constant) are jointly zero.
    """
    output = {}

    include_dl = True if second_stage_spec == 't4' else False
    curr_dict = load_data.create_curr_df(df, y_variables_cp, x_variables_cp, countries, include_dl=include_dl)
    gammas = curr_dict['gammas']
    df = curr_dict['df']

    if second_stage_spec == 't4':
        df['dl_cp_i'] = df['dl_fwd_i'] * df['cp_factor_i']
        df['dl_cp_j'] = df['dl_fwd_j'] * df['cp_factor_j']

        formula = 'curr_ret_fwd ~ dl_fwd_i + dl_cp_i + dl_fwd_j + dl_cp_j'
        g_fun = g_fun_t4

        h_T = np.hstack([np.ones([df.shape[0], 1]), df[['dl_fwd_i', 'dl_cp_i', 'dl_fwd_j', 'dl_cp_j']].values])

    elif second_stage_spec == 't5':
        df['y_dif'] = df['y1_j'] - df['y1_i']

        formula = 'curr_ret_fwd ~ cp_factor_i + cp_factor_j + y_dif'
        g_fun = g_fun_t5
        h_T = np.hstack([np.ones([df.shape[0], 1]), df[['cp_factor_i', 'cp_factor_j', 'y_dif']].values])

    ols = sm.ols(formula, data=df).fit()
    params2stage = ols.params
    res2stage = ols.resid

    output['params'] = ols.params.to_frame(name='params')
    output['rsquared_adj'] = ols.rsquared_adj

    if not include_se:
        return output

    # 3. Calculate GMM standard errors. See Appendix C of Hodrick-Tomunen (2019) for formulas.
    f_T_i = np.hstack([np.ones([df.shape[0], 1]), df[[i + '_i' for i in x_variables_cp]].values])
    f_T_j = np.hstack([np.ones([df.shape[0], 1]), df[[i + '_j' for i in x_variables_cp]].values])

    T = f_T_i.shape[0]
    N_x = f_T_i.shape[1]
    N_h = h_T.shape[1]

    theta = np.append(gammas[countries[1]].values, gammas[countries[0]].values)
    theta = np.append(theta, params2stage.values)

    EffT_i = (f_T_i.T @ f_T_i) / T
    EffT_j = (f_T_j.T @ f_T_j) / T

    d_mat1 = np.hstack([-EffT_i, np.zeros([f_T_i.shape[1], theta.shape[0] - N_x])])
    d_mat2 = np.hstack([np.zeros([f_T_j.shape[1], N_x]), -EffT_j, np.zeros([f_T_i.shape[1], theta.shape[0] - 2 * N_x])])
    d_mat3 = nd.Jacobian(g_fun, order=6)(theta, f_T_i, f_T_j, df)

    d_mat = np.vstack([d_mat1, d_mat2, d_mat3])

    u1 = np.multiply(np.tile(df['resid_1stage_i'].values, [N_x, 1]).T, f_T_i)
    u2 = np.multiply(np.tile(df['resid_1stage_j'].values, [N_x, 1]).T, f_T_j)
    u3 = np.multiply(np.tile(res2stage, [N_h, 1]).T, h_T)

    u = np.hstack([u1, u2, u3])

    S = S_matrix(u, 18, newey_west=True, demean=True)
    VCV = np.linalg.inv(d_mat) @ S @ np.linalg.inv(d_mat).T / T
    stderrs = np.sqrt(np.diag(VCV))
    output['params']['se'] = stderrs[N_x * 2:]

    # 4. Wald tests.
    if second_stage_spec == 't5':
        ix = int(N_x*2+1)

        R_cp = np.hstack([np.zeros([2, ix]), np.eye(2), np.zeros([2, 1])])
        r_cp = np.array([0, 0])

        R_uip = np.array([[0] * (ix + 2) + [1]])

        r_uip0 = 0
        r_uip1 = 1

        for name, (R, r) in {'cp': (R_cp, r_cp), 'uip0': (R_uip, r_uip0), 'uip1': (R_uip, r_uip1)}.items():
            chi2_stat = np.asscalar((R @ theta - r).T @ np.linalg.inv((R @ VCV @ R.T)) @ (R @ theta - r))

            dof = R.shape[0]

            output['wald_test_{}'.format(name)] = {'value': chi2_stat}
            output['wald_test_{}'.format(name)]['p_value'] = 1 - chi2.cdf(chi2_stat, dof)
            output['wald_test_{}'.format(name)]['dof'] = dof

    return output


def g_fun_t4(x, f_T_i, f_T_j, df):
    gamma_i = np.matrix(x[0:f_T_i.shape[1]]).T
    gamma_j = np.matrix(x[f_T_i.shape[1]:f_T_i.shape[1] + f_T_j.shape[1]]).T
    theta2stage = np.matrix(x[f_T_i.shape[1] + f_T_j.shape[1]:]).T

    cp_factor_i = gamma_i.T @ f_T_i.T
    cp_factor_j = gamma_j.T @ f_T_j.T

    df_i_fwd = df[['dl_fwd_i']].values
    df_j_fwd = df[['dl_fwd_j']].values

    h_T = np.hstack([np.ones([df_i_fwd.shape[0], 1]), df_i_fwd, np.multiply(df_i_fwd, cp_factor_i.T),
                     df_j_fwd, np.multiply(df_j_fwd, cp_factor_j.T)])

    y_T = df[['curr_ret_fwd']].values
    err_s = y_T.T - theta2stage.T @ h_T.T

    return (err_s @ h_T / err_s.shape[1]).T


def g_fun_t5(x, f_T_i, f_T_j, df):
    gamma_i = np.matrix(x[0:f_T_i.shape[1]]).T
    gamma_j = np.matrix(x[f_T_i.shape[1]:f_T_i.shape[1] + f_T_j.shape[1]]).T
    theta2stage = np.matrix(x[f_T_i.shape[1] + f_T_j.shape[1]:]).T

    cp_factor_i = gamma_i.T @ f_T_i.T
    cp_factor_j = gamma_j.T @ f_T_j.T

    h_T = np.hstack([np.ones([cp_factor_i.T.shape[0], 1]), cp_factor_i.T, cp_factor_j.T, df[['y_dif']].values])

    y_T = df[['curr_ret_fwd']].values
    err_s = y_T.T - theta2stage.T @ h_T.T

    return (err_s @ h_T / err_s.shape[1]).T
