import numpy as np
import pandas as pd
from pandas.tseries.offsets import MonthEnd

from modules.analyses import cp2stage, cp_freeconstant


def run(df, y_variables, x_variables, oos_startdate, model):
    """
    Run two-stage Cochrane-Piazzesi regressions or free constant model recursively, and return a panel of parameter
    estimates and various out-of-sample statistics.

    Parameters
    ----------
    df: pandas.DataFrame
        A dataframe that contains time series for all excess bond returns and forward spreads (for one country).
    y_variables: list
        A list of df column names that contain excess bond returns used in the analysis.
    x_variables: list
        A list of df column names that contain forward spreads used in the analysis.
    oos_startdate: string
        Start date of out-of-sample period.
    model: {'cp', 'cp_freeconstant'}
        If 'cp', run standard Cochrane-Piazzesi model.
        If 'cp_freeconstant', run a free constant model described in Section 1 of Online Appendix.

    Output
    ------
    A dictionary with following items:

    rsquared: numpy.ndarray
        Out-of-sample R-squared statistics.
    mse_f: numpy.ndarray
        Clark and McCracken (2005) MSE-F statistics.
    dif_se: pandas.DataFrame
        Difference in squared prediction errors between historical mean and model.
    gammas: pandas.DataFrame
        Gamma estimates for each out-of-sample date.
    bs: pandas.DataFrame
        Second-stage parameter estimates and their GMM standard errors.
    """
    assert model in ['cp', 'cp_freeconstant', 'constrained_ols']
    output = {}
    df = df.copy()
    df['ones'] = 1

    pred_model = []
    pred_mean = []
    t_gamma = []
    t_b = []
    dates = df.loc[df['date'] >= oos_startdate, 'date'].unique()
    T_oos = len(dates)

    for t in dates:
        # Next row: data available for parameter estimation ends 12 months before OOS prediction, because y-variable
        # is forward return over 12 months.
        df2 = df[(df['date'] + MonthEnd(12)) <= t].copy()

        if model == 'cp':
            cp_result = cp2stage.run(df2, y_variables, x_variables)
            cp_factor_pred = np.dot(df.loc[(df['date'] == t), ['ones'] + x_variables].values,
                                    cp_result['gammas']['params'].values)
            pred_model += [[t] + list(cp_result['bs']['params'].values * cp_factor_pred)]

        elif model == 'cp_freeconstant':
            cp_result = cp_freeconstant.run(df2, y_variables, x_variables, include_se=False)

            alphas = cp_result['alphas']['params'].values
            deltas = cp_result['deltas']['params'].values
            gammas = cp_result['gammas']['params'].values
            bs = cp_result['bs']['params'].values

            cp_factor_pred = ((np.matrix(bs).T @ (np.matrix(gammas) @ (df.loc[df['date'] == t, x_variables].values
                                                                       - deltas).T)).T + alphas)

            pred_model += [[t] + np.squeeze(np.array(cp_factor_pred)).tolist()]

        pred_mean += [[t] + df2[y_variables].mean().tolist()]

        t_gamma += [[t] + cp_result['gammas']['params'].tolist()]
        t_b += [[t] + cp_result['bs']['params'].tolist()]

    df = df.merge(pd.DataFrame(pred_model, columns=['date'] + y_variables),
                  on='date', how='inner', validate='1:1', suffixes=['', '_model_pred'])
    df = df.merge(pd.DataFrame(pred_mean, columns=['date'] + y_variables),
                  on='date', how='inner', validate='1:1', suffixes=['', '_hm_pred'])

    se_model = np.power((df[y_variables].values - df[_add_suffix(y_variables, '_model_pred')].values), 2)
    se_hm = np.power((df[y_variables].values - df[_add_suffix(y_variables, '_hm_pred')].values), 2)

    mse_model = se_model.mean(axis=0)
    mse_hm = se_hm.mean(axis=0)

    output['rsquared'] = 1 - mse_model / mse_hm
    output['mse_f'] = T_oos * (mse_hm - mse_model) / mse_model

    assert se_hm.shape == se_model.shape
    output['dif_se'] = pd.DataFrame(se_hm - se_model)
    output['dif_se'].columns = y_variables
    output['dif_se'].insert(0, 'date', dates)

    cols = ['date'] + ['k={}'.format(i) for i in range(1, len(t_gamma[0]))]
    output['gammas'] = pd.DataFrame(t_gamma, columns=cols).set_index('date')

    cols = ['date'] + ['k={}'.format(i) for i in range(1, len(t_b[0]))]
    output['bs'] = pd.DataFrame(t_b, columns=cols).set_index('date')

    return output


def _add_suffix(str_list, suffix):
    return [j + suffix for j in str_list]
