import numpy as np
import pandas as pd
from pandas.tseries.offsets import MonthEnd

from modules.analyses import cp2stage


def run(df, y_variables, x_variables, oos_startdate):
    """
    Run constrained Cochrane-Piazzesi model recursively (parameters of all countries are the same),
    and return a panel of parameter estimates and various out-of-sample statistics.

    Parameters
    ----------
    df: pandas.DataFrame
        A dataframe that contains time series for all excess bond returns and forward spreads (for all countries).
        Must be a balanced panel.
    y_variables: list
        A list of df column names that contain excess bond returns used in the analysis.
    x_variables: list
        A list of df column names that contain forward spreads used in the analysis.
    oos_startdate: string
        Start date of out-of-sample period.

    Output
    ------
    A dictionary with following items:

    rsquared: numpy.ndarray
        Out-of-sample R-squared statistics.
    mse_f: numpy.ndarray
        Clark and McCracken (2005) MSE-F statistics.
    dif_se: pandas.DataFrame
        Difference in squared prediction errors between historical mean and model.
    """
    countries = df['country'].unique()
    output = {}

    assert (df.groupby('country')['date'].agg(['min', 'max', 'count']).nunique() == 1).all(), 'Panel must be balanced.'
    df = df.copy()

    pred_model = {}
    pred_mean = {}
    for i in countries:
        pred_model[i] = []
        pred_mean[i] = []

    dates = df.loc[df['date'] >= oos_startdate, 'date'].unique()
    T_oos = len(dates)

    for t in dates:
        # Next row: data available for parameter estimation ends 12 months before OOS prediction, because y-variable
        # is return over 12 months.
        df2 = df[(df['date'] + MonthEnd(12)) <= t].copy()

        # Demean x and y-variables.
        xmeans = df2.groupby('country', as_index=False)[x_variables].mean()
        ymeans = df2.groupby('country', as_index=False)[y_variables].mean()
        df2 = df2.merge(xmeans, on='country', suffixes=['', '_mean'])
        df2 = df2.merge(ymeans, on='country', suffixes=['', '_mean'])
        for i in countries:
            pred_mean[i] += [[t] + ymeans.loc[ymeans['country'] == i, y_variables].values.reshape(-1).tolist()]
        for v in x_variables + y_variables:
            df2[v] = df2[v] - df2[v + '_mean']
            df2 = df2.drop(columns=[v + '_mean'])

        cp_result = cp2stage.run(df2, y_variables, x_variables, include_constant_1stage=False)

        gammas = cp_result['gammas']['params'].values
        bs = cp_result['bs']['params'].values

        for i in countries:
            x_vals = df.loc[(df['date'] == t) & (df['country'] == i), x_variables].values
            cp_factor_pred = ((np.matrix(bs).T @ (np.matrix(gammas) @ x_vals.T)).T
                              + ymeans.loc[ymeans['country'] == i, y_variables].values)

            pred_model[i] += [[t] + np.squeeze(np.array(cp_factor_pred)).tolist()]

    for i in countries:
        output[i] = {}
        df2 = df[df['country'] == i].copy()
        df2 = df2.merge(pd.DataFrame(pred_model[i], columns=['date'] + y_variables),
                        on='date', how='inner', validate='1:1', suffixes=['', '_model_pred'])
        df2 = df2.merge(pd.DataFrame(pred_mean[i], columns=['date'] + y_variables),
                        on='date', how='inner', validate='1:1', suffixes=['', '_hm_pred'])

        se_model = np.power((df2[y_variables].values - df2[_add_suffix(y_variables, '_model_pred')].values), 2)
        se_hm = np.power((df2[y_variables].values - df2[_add_suffix(y_variables, '_hm_pred')].values), 2)

        mse_model = se_model.mean(axis=0)
        mse_hm = se_hm.mean(axis=0)

        output[i]['rsquared'] = 1 - mse_model / mse_hm
        output[i]['mse_f'] = T_oos * (mse_hm - mse_model) / mse_model

        assert se_hm.shape == se_model.shape
        output[i]['dif_se'] = pd.DataFrame(se_hm - se_model)
        output[i]['dif_se'].columns = y_variables
        output[i]['dif_se'].insert(0, 'date', dates)

    return output


def _add_suffix(str_list, suffix):
    return [j + suffix for j in str_list]
