import numpy as np
import pandas as pd
from pandas.tseries.offsets import MonthEnd

from modules.analyses import curr2stage
from modules.load_data import create_curr_df


def run(df, y_variables_cp, x_variables_cp, oos_startdate, countries):
    """
    Run two-stage regressions to explain currency market excess returns with CP factors recursively,
    and return a panel of parameter estimates and various out-of-sample statistics.


    Parameters
    ----------
    df: pandas.DataFrame
        A dataframe that contains time series for all excess bond returns and forward spreads.
        (for all countries in 'countries').
    y_variables: list
        A list of df column names that contain excess bond returns used in the analysis.
    x_variables: list
        A list of df column names that contain forward spreads used in the analysis.
    oos_startdate: string
        Start date of out-of-sample period.
    countries: list
        Defines the currency pair for which the analysis is run. Format: [target_currency (j), base_currency (i)].
        Example: ['GBP', 'USD'].

    Output
    ------
    A dictionary with following items:

    rsquared: numpy.float64
        Out-of-sample R-squared statistic.
    mse_f: numpy.float64
        Clark and McCracken (2005) MSE-F statistic.
    """
    output = {}
    df = df.copy()

    pred_model = []
    pred_mean = []

    dates = df.loc[df['date'] >= oos_startdate, 'date'].unique()
    T_OOS = len(dates)

    dfc = create_curr_df(df, y_variables_cp, x_variables_cp, countries)['df']

    dfc['ones'] = 1

    dfc['y_dif'] = dfc['y1_j'] - dfc['y1_i']
    x_cols = ['ones', 'cp_factor_i', 'cp_factor_j', 'y_dif']

    for t in dates:
        # Next row: data available for parameter estimation ends 12 months before OOS prediction, because y-variable
        # is return over 12 months.
        df2 = df[(df['date'] + MonthEnd(12)) <= t].copy()

        curr_result = curr2stage.run(df2, y_variables_cp, x_variables_cp, 't5',
                                     countries, include_se=False)

        pred = np.dot(dfc.loc[(dfc['date'] == t), x_cols].values,
                      curr_result['params']['params'].values)

        pred_model += [[t] + pred.tolist()]

        dfc2 = dfc[(dfc['date'] + MonthEnd(12)) <= t].copy()
        pred_mean += [[t, dfc2['curr_ret_fwd'].mean()]]

    dfc = dfc.merge(pd.DataFrame(pred_model, columns=['date', 'pred_model']),
                    on='date', how='inner', validate='1:1')
    dfc = dfc.merge(pd.DataFrame(pred_mean, columns=['date', 'pred_hm']),
                    on='date', how='inner', validate='1:1')

    MSE_MODEL = np.power((dfc['curr_ret_fwd'].values - dfc['pred_model'].values), 2).mean(axis=0)
    MSE_HM = np.power((dfc['curr_ret_fwd'].values - dfc['pred_hm'].values), 2).mean(axis=0)

    output['rsquared'] = 1 - MSE_MODEL / MSE_HM
    output['mse_f'] = T_OOS * (MSE_HM - MSE_MODEL) / MSE_MODEL

    return output
