import numpy as np
from scipy.stats import chi2


def run(cp_result1, cp_result2):
    """
    Test the equality of b and gamma parameters given two outputs from cp2stage.run function.

    Parameters
    ----------
    cp_result1: dictionary
        First input from cp2stage.run
    cp_result2: dictionary
        Second input from cp2stage.run

    Output
    ------
    A dictionary with following items:

    gamma_chi2: float
        Chi2-stat from testing the equality of gammas.
    gamma_p_value: numpy.float64
        P-value for gamma_chi2.
    b_chi2: float
        Chi2-stat from testing the equality of b's.
    b_p_value: numpy.float64
        P-value for b_chi2.
    dofs: list
        Degrees of freedom for chi2-stats (first element b, second gamma).
    """
    output = {}

    gamma_dif = np.matrix((cp_result1['gammas']['params'] - cp_result2['gammas']['params']).values).T
    b_dif = np.matrix((cp_result1['bs']['params'] - cp_result2['bs']['params']).values[:-1]).T

    N_g = gamma_dif.shape[0]
    N_b = b_dif.shape[0]

    gamma_vcv = (cp_result1['VCV'] + cp_result2['VCV'])[:N_g, :N_g]
    b_vcv = (cp_result1['VCV'] + cp_result2['VCV'])[N_g:-1, N_g:-1]

    output['b_chi2'] = np.asscalar(b_dif.T @ np.linalg.inv(b_vcv) @ b_dif)
    output['b_p_value'] = 1 - chi2.cdf(output['b_chi2'], N_b)

    output['gamma_chi2'] = np.asscalar(gamma_dif.T @ np.linalg.inv(gamma_vcv) @ gamma_dif)
    output['gamma_p_value'] = 1 - chi2.cdf(output['gamma_chi2'], N_g)

    output['dofs'] = [N_b, N_g]

    return output
