import numpy as np
import pandas as pd
from sklearn.decomposition import PCA

from data.create_us import create_us
from modules.analyses import cp2stage
from modules.utils.data_utils import check_ts_integrity, check_primary_key_integrity

YIELD_NAMES = ['USD', 'GBP', 'EUR', 'CHF', 'CAD', 'JPY', 'AUD', 'SEK', 'NOK']

START_DATE = pd.to_datetime('10/31/1973')
END_DATE = pd.to_datetime('12/31/2016')


def load():
    """
    Load yields for all currencies and maturities from csv files,
    create all price variables described in Section 2 of the paper, and return everything as a dataframe.
    Primary key is ['date', 'country'].
    """
    df = pd.DataFrame()
    for i in YIELD_NAMES:
        if i == 'USD':
            try:
                df2 = pd.read_csv('data/{}.csv'.format(i), parse_dates=['date'])
            except FileNotFoundError:
                df2 = create_us(save=False)
                df2.to_csv('data/USD.csv', index=False)
        else:
            df2 = pd.read_csv('data/{}.csv'.format(i), parse_dates=['date'])
        # Rename yield columns to y1 - y5
        old_names = ['{}{}Y'.format(i, n) for n in range(1, 6)]
        new_names = ['y1', 'y2', 'y3', 'y4', 'y5']
        df2 = df2.rename(columns=dict(zip(old_names, new_names)))

        df2.insert(1, 'country', i)
        df2 = df2[(df2['date'] >= START_DATE) & (df2['date'] <= END_DATE)]
        assert check_ts_integrity(df2['date'], 'monthly')

        df = df.append(df2, ignore_index=True)

    for n in range(1, 6):
        df['p{}'.format(n)] = - df['y{}'.format(n)] * n

    for n in range(2, 6):
        df['f{}'.format(n)] = df['p{}'.format(n-1)] - df['p{}'.format(n)]
        df['fs{}'.format(n)] = df['f{}'.format(n)] - df['y1']
        df['fs{}_bar'.format(n)] = (df['fs{}'.format(n)]
                                    + df.groupby('country')['fs{}'.format(n)].shift(1)
                                    + df.groupby('country')['fs{}'.format(n)].shift(2)) / 3
    for n in range(2, 6):
        df['r{}_fwd'.format(n)] = df.groupby('country')['p{}'.format(n-1)].shift(-12) - df['p{}'.format(n)]
        df['rx{}_fwd'.format(n)] = df['r{}_fwd'.format(n)] - df['y1']

    df = df.dropna(how='any')  # Drops first 2 obs (because fs_bar is missing) and last 12 (because rx_fwd is missing).
    check_primary_key_integrity(df, ['date', 'country'])

    return df


def load_fx_ret(target_curr, base_curr):
    """
    Load exchange rate data and calculate the continuously compounded rate of appreciation of base currency i
    relative to target currency j during a year. This is denoted with ∆s_ij,t+1 in the paper.
    Return a data frame.
    """
    assert target_curr in YIELD_NAMES, 'target_curr must be in {}.'.format(YIELD_NAMES)
    assert base_curr in YIELD_NAMES,  'base_curr must be in {}.'.format(YIELD_NAMES)
    assert target_curr != base_curr, 'target_curr and base_curr must be different.'

    df = pd.read_csv('data/FX_to_USD.csv', parse_dates=['date']).sort_values('date').reset_index(drop=True)
    df = df[(df['date'] >= START_DATE) & (df['date'] <= END_DATE)]

    assert check_ts_integrity(df['date'], 'monthly')

    if base_curr == 'USD':
        df['cur'] = np.log(df[target_curr + 'USD'])
    elif target_curr == 'USD':
        df['cur'] = np.log(1 / df[base_curr + 'USD'])

    elif base_curr != 'USD':
        df['cur'] = np.log(df[target_curr + 'USD'] / df[base_curr + 'USD'])

    df['s_fwd'] = (df['cur'].shift(-12) - df['cur']) * 100
    df = df.dropna(how='any')  # Drops last 12 observations.

    return df[['date', 's_fwd']]


def create_curr_df(df, y_variables_cp, x_variables_cp, countries, include_dl=False):
    """
    Create a data frame for currency analyses. Primary key is 'date'.

    Parameters
    ----------
    df: pandas.DataFrame
        A dataframe that contains time series for all excess bond returns and forward spreads.
    y_variables_cp: list
        A list of df column names that contain excess bond returns used to construct CP factor.
    x_variables_cp: list
        A list of df column names that contain forward spreads used to construct CP factor.
    countries: list
        Defines the currency pair for which the analysis is run. Format: [target_currency (j), base_currency (i)].
        Example: ['GBP', 'USD'].
    include_df: boolean, default False
        If True, include changes in the first principal components of the yields of base and target countries as
        columns to output.

    Output
    ------
    A dictionary with following items
    df: pandas.DataFrame
        Contains forward spreads, CP factors and their residuals of base and target countries,
        and excess rate of return on currency markets (∆s_ij + r_j -r_i).
    gammas: dictionary
        Contains CP first-stage gamma estimates for base and target countries.
    """
    output = {}

    df = df[df['country'].isin(countries)].copy()
    assert len(df['country'].unique()) == 2, "Input 'countries' should have exactly two valid currencies as elements."

    output['gammas'] = {}
    df2 = pd.DataFrame()
    for i in countries:  # First stages
        df_temp = df[(df['country'] == i)].copy()
        cp_result = cp2stage.run(df_temp, y_variables_cp, x_variables_cp)
        cp_result['1stage']['country'] = i

        output['gammas'][i] = cp_result['gammas']['params']
        df2 = df2.append(cp_result['1stage'])

    df = df[['date', 'country', 'y1', 'y2', 'y3', 'y4', 'y5'] + x_variables_cp]

    if include_dl:
        df = _add_dl(df, countries)

    df = df.drop(columns=['y2', 'y3', 'y4', 'y5'])

    df = df.merge(df2, on=['date', 'country'], validate='1:1')

    df = df.pivot('date', 'country')
    df.columns = df.columns.map('_'.join)
    df.columns = df.columns.str.replace(countries[0], 'j')
    df.columns = df.columns.str.replace(countries[1], 'i')

    df = df.reset_index().merge(load_fx_ret(countries[0], countries[1]), on='date', validate='1:1')

    df['curr_ret_fwd'] = (-df['s_fwd'] + df['y1_j'] - df['y1_i'])
    df = df.dropna(how='any').drop(columns=['s_fwd'])

    output['df'] = df

    return output


def _add_dl(df, countries):
    """Calculate changes in the first principal components of the yields and add to dataframe."""
    df3 = pd.DataFrame()
    for i in countries:
        df2 = df[(df['country'] == i)].copy().sort_values('date').reset_index(drop=True)
        assert check_ts_integrity(df2['date'], 'monthly')

        df2['l'] = PCA(n_components=1).fit_transform(df2[['y1', 'y2', 'y3', 'y4', 'y5']].values)

        df2['dl_fwd'] = (df2['l'].shift(-12) - df2['l'])
        df3 = df3.append(df2[['date', 'country', 'dl_fwd']])

    df = df.merge(df3, on=['date', 'country'], validate='1:1')

    return df
