import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from prettytable import PrettyTable

from modules import load_data
from modules.analyses import cp2stage, oos
from modules.utils.table import Table


def create(output='console'):
    OOS_START = '12/31/2003'

    Y_VARIABLES = ['rx2_fwd', 'rx3_fwd', 'rx4_fwd', 'rx5_fwd']
    X_VARIABLES = ['fs2_bar', 'fs3_bar', 'fs4_bar', 'fs5_bar']

    df = load_data.load()

    tdata = []
    se_dict = {}
    for i in df['country'].unique():
        df2 = df[df['country'] == i].copy()
        oos_result = oos.run(df2, Y_VARIABLES, X_VARIABLES, OOS_START, 'cp')

        tdata += [[i] + list(oos_result['rsquared']) + list(oos_result['mse_f'])]
        _print_params(oos_result['gammas'], oos_result['bs'], i, OOS_START, output, freeconst=False)
        se_dict[i] = oos_result['dif_se']

    tdata2 = []
    for i in df['country'].unique():
        df2 = df[df['country'] == i].copy()
        oos_result = oos.run(df2, Y_VARIABLES, X_VARIABLES, OOS_START, 'cp_freeconstant')

        tdata2 += [[i] + list(oos_result['rsquared']) + list(oos_result['mse_f'])]
        _print_params(oos_result['gammas'], oos_result['bs'], i, OOS_START, output, freeconst=True)

    _fig1(se_dict, Y_VARIABLES, X_VARIABLES, output)

    print('\nTable 6')
    if output == 'console':
        for n, td in enumerate([tdata, tdata2]):
            table = PrettyTable(['CUR', 'R2(rx2)', 'R2(rx3)', 'R2(rx4)', 'R2(rx5)',
                                 'MSE-F(rx2)', 'MSE-F(rx3)', 'MSE-F(rx4)', 'MSE-F(rx5)'])
            table.float_format = '0.2'
            for i in td:
                table.add_row(i)
            print(table)

    elif output == 'file':
        colnames = ['CUR'] + ['$rx_{{t+1}}^{{({})}}$'.format(i) for i in range(2, 6)] * 2

        justs = ['l'] + ['r'] * (len(colnames) - 1)

        table = Table(len(colnames), justs=justs, label='OoS_forecasts_CP',
                      caption='Out-of-Sample Forecasts for Excess Bond Returns from Cochrane-Piazzesi Models',
                      caption_position='bottom')
        # The following panel header must be added manually to the .tex output file:
        # 'Panel A: Basic Models vs. Historical Means'
        table.add_panel(pd.DataFrame(tdata),
                        supheaders=['', '$R^2$', 'MSE-F'],
                        supheader_cols=[1, (2, 5), (6, 9)],
                        headers=colnames,
                        formatting='%.2f')

        table.add_panel(pd.DataFrame(tdata2),
                        headers=['Panel B: Models With Free Constants vs. Historical Means'],
                        header_cols=[(1, 9)],
                        formatting='%.2f')

        table.print_table(fp='output/tables/OoS_forecasts_CP.tex')

    pass


def _print_params(dfg, dfb, country, oos_startdate, output, freeconst=False):
    """
    Print a time series graph of gammas and bs from recursive estimations.

    Parameters
    ----------
    dfg: pandas.DataFrame
        A dataframe that contains time series for gamma estimates.
    dfb: pandas.DataFrame
        A dataframe that contains time series for b estimates.
    country: string
        Name of the country that is being printed.
    oos_startdate: string
        Start date of out-of-sample period.
    output: {'console', 'file'}
        Determines whether output is printed to console or to an eps file.
    freeconst: boolean, default False
        If True, add '_freeconst' suffix to the file name (only relevant if output='file').
    """
    fig, ax = plt.subplots(nrows=2, ncols=1, figsize=(10, 7))

    dfg.plot(ax=ax[0])
    ax[0].set_xlabel('')
    ax[0].set_ylabel('$\\gamma_k$', fontsize=14)
    ax[0].set_xlim([oos_startdate, dfb.index.max().strftime('%m/%d/%Y')])
    ax[0].legend(loc='lower right', fontsize=14)
    ax[0].tick_params(axis='both', labelsize=14)

    dfb.plot(ax=ax[1])
    ax[1].set_xlabel('')
    ax[1].set_ylabel('$b_k$', fontsize=14)
    ax[1].set_xlim([oos_startdate, dfb.index.max().strftime('%m/%d/%Y')])
    ax[1].set_ylim((0, dfb.max().max()*1.2))
    ax[1].legend(loc='lower right', fontsize=14)
    ax[1].tick_params(axis='both', labelsize=14)

    folder = 'freeconst' if freeconst else 'cp'

    print('\nFig 2 ({} {})'.format(country, folder))
    if output == 'file':
        plt.savefig('output/figures/{}/OosParams{}.eps'.format(folder, country), format='eps')
    plt.close()
    pass


def _fig1(se_dict, y_variables, x_variables, output):
    is_dict = cp2stage.calc_dif_se(y_variables, x_variables)
    SE_STARTDATE = '12/31/2006'
    COUNTRY_GROUPS = [['USD', 'EUR', 'JPY'],
                      ['GBP', 'CHF', 'CAD'],
                      ['AUD', 'SEK', 'NOK']]
    for col in se_dict['USD'].columns[1:]:
        for cg in COUNTRY_GROUPS:
            fig, ax = plt.subplots(nrows=3, ncols=1, figsize=(10, 12))
            for n, i in enumerate(cg):
                df = se_dict[i].set_index('date')[[col]]
                dfi = is_dict[i]
                dfi = dfi[(dfi['date'] >= df.index.min()) & (dfi['date'] <= df.index.max())]
                se_dates = df.index[df.index >= SE_STARTDATE].unique()

                df['se'] = 0
                for t in se_dates:
                    d = df.loc[df.index <= t, col].values
                    T = d.shape[0]
                    h = 12
                    se_d = _se_d(d, h)
                    se_d_adj = se_d / np.sqrt((T + 1 - 2*h + (1/T)*h*(h-1))/T)
                    se_sse = se_d_adj * T
                    df.loc[df.index == t, 'se'] = se_sse
                df['sse'] = df[col].cumsum()
                df['upper'] = df['sse'] + 2 * df['se']
                df['lower'] = df['sse'] - 2 * df['se']

                df['sse'].plot(ax=ax[n], color='k')
                ax[n].fill_between(df.index, df['upper'].values, df['lower'].values, facecolor='gray', alpha=0.1,
                                   linewidth=0.0)

                dfi.set_index('date')[col].cumsum().plot(ax=ax[n], color='k', style=':', linewidth=1)
                ax[n].set_title(i, fontweight='bold', fontsize=14)
                ax[n].set_xlim([df.index.min().strftime('%m/%d/%Y'), df.index.max().strftime('%m/%d/%Y')])
                ax[n].axhline(y=0, linewidth=1, color='k')
                ax[n].set_xlabel('')
                ax[n].set_ylabel('Cumulative\nSSE Difference', fontsize=14)
                ax[n].tick_params(axis='both', labelsize=14)

            print('\nFig 1({})'.format(col))
            if output == 'file':
                plt.savefig('output/figures/goyal_welch/{}.eps'.format(col + '_'.join(cg)), format='eps')
            plt.close()

    pass


def _se_d(series, lags):
    """Calculate Diebold-Mariano standard error."""
    T = series.shape[0]

    S = np.asscalar(np.cov(series))

    for k in range(1, lags):
        S += 2 * np.cov(series[k:],  series[:-k])[1, 0]

    # If the estimate for S is negative, return zero. Figure won't show any bands for that date.
    return np.sqrt(max(S / T, 0))
