import warnings

import pandas as pd
from prettytable import PrettyTable

from modules import load_data
from modules.analyses import cp2stage, test_param_equality, unrestricted_variance_explained
from modules.utils.table import Table


def create(output='console'):
    SUBSAMPLES = [['12/31/1973', '12/31/2002'],
                  ['12/31/2003', '12/31/2015']]

    Y_VARIABLES = ['rx2_fwd', 'rx3_fwd', 'rx4_fwd', 'rx5_fwd']
    X_VARIABLES = ['fs2_bar', 'fs3_bar', 'fs4_bar', 'fs5_bar']

    df = load_data.load()

    tdata = []
    tdata2 = []
    for i in df['country'].unique():
        for n, j in enumerate(SUBSAMPLES):
            df_temp = df[(df['country'] == i) & (df['date'] >= j[0]) & (df['date'] <= j[1])].copy()

            cp_result = cp2stage.run(df_temp, Y_VARIABLES, X_VARIABLES)

            pc1 = unrestricted_variance_explained.run(df_temp, Y_VARIABLES, X_VARIABLES, n_components=1)

            tdata += [[i + str(n+1)]
                      + cp_result['bs']['params'].tolist()[:-1]
                      + cp_result['gammas']['params'].tolist()
                      + [cp_result['wald_test']['value'], cp_result['rsquared_adj'], pc1]]

            tdata += [['']
                      + cp_result['bs']['se'].tolist()[:-1]
                      + cp_result['gammas']['se'].tolist()
                      + [cp_result['wald_test']['p_value'], '', '']]

            if n == 0:
                cp_result1 = cp_result.copy()

        if len(SUBSAMPLES) == 2:
            equality_result = test_param_equality.run(cp_result1, cp_result)

            tdata2 += [[i, equality_result['b_chi2'], equality_result['gamma_chi2']]]
            tdata2 += [['',  equality_result['b_p_value'], equality_result['gamma_p_value']]]
            dofs = equality_result['dofs']

    _print_table1(tdata, output=output)

    if len(SUBSAMPLES) == 2:
        _print_table2(tdata2, dofs, output=output)
    elif output == 'file':
        warnings.warn('Other than two subsamples were defined. Table 2 was not updated.')

    pass


def _print_table1(tdata, output='console'):
    """
    Print Table 1.

    Parameters
    ----------
    tdata: list
        A list of list containing Table 1 data. Each inner list is a row with following format for
        two consecutive rows:
            [Currency Name, b2, b3, b4, gamma1, gamma2, gamma3, gamma4, gamma5, chi2, R2, %PC1],
            ['', b2_se, b3_se, b4_se, gamma1_se, gamma2_se, gamma3_se, gamma4_se, gamma5_se, chi2_p-value, '', '']
    output: {'console', 'file'}, default 'console'
        Output location.
    """
    print('\nTable 1')
    if output == 'console':
        table = PrettyTable(['CUR', 'b2', 'b3', 'b4', 'gamma1', 'gamma2', 'gamma3',
                             'gamma4', 'gamma5', 'chi2', 'R-squared', 'PC1'])
        table.float_format = '0.2'
        for i in tdata:
            table.add_row(i)
        print(table)

    elif output == 'file':
        colnames = ['CUR', '$b_2$', '$b_3$', '$b_4$', '$\\gamma_1$', '$\\gamma_2$', '$\\gamma_3$',
                    '$\\gamma_4$', '$\\gamma_5$', '$\\chi^2(4)$', '$R^2$', '\\%PC1']
        justs = ['l'] + ['d'] * (len(colnames) - 1)
        formatting = ['%.2f'] * (len(colnames) - 1) + ['%.3f']
        brackets = ['('] * 9 + ['angled'] * 3

        table = Table(12, justs=justs, label='SingleFactorModelPre_Post',
                      caption='The Single Factor Cochrane and Piazzesi (2005) Model', caption_position='bottom')

        for row1, row2, row3, row4 in zip(tdata[0::4], tdata[1::4], tdata[2::4], tdata[3::4]):
            table.add_panel(pd.DataFrame([row1, row2, row3, row4]), headers=colnames,
                            formatting=formatting,
                            regression_specs={'stars': None, 'dof': None, 'means': None, 'bracket_marker': brackets})
            colnames = None

        table.print_table(fp='output/tables/SingleFactorModelPre_Post.tex')

    pass


def _print_table2(tdata, dofs, output='console'):
    """
    Print Table 2.

    Parameters
    ----------
    tdata: list
        A list of list containing Table 2 data. Each inner list is a row with following format for
        two consecutive rows:
            [Currency Name, chi2(b), chi2(gamma)],
            ['', chi2(b)_p-value, chi2(gamma)_p-value]
    dofs: list
        Degrees of freedom for chi2(b) and chi2(gamma).
    output: {'console', 'file'}, default 'console'
        Output location.
    """
    print('\nTable 2')
    if output == 'console':
        table = PrettyTable(['CUR', 'chi2(b)', 'chi2(gamma)'])
        table.float_format = '0.2'
        for i in tdata:
            table.add_row(i)
        print(table)

    elif output == 'file':
        colnames = ['CUR',
                    "$\\chi^2 ({})$ for $\\mathbf{{b}}$'s".format(dofs[0]),
                    "$\\chi^2 ({})$ for $\\vgamma$'s".format(dofs[1])]

        justs = ['l'] + ['d'] * (len(colnames) - 1)

        table = Table(3, justs=justs, label='tests_of_equality',
                      caption='Tests of Equality of Coefficients for the Two Samples', caption_position='bottom')
        table.add_panel(pd.DataFrame(tdata), headers=colnames,
                        formatting='%.2f',
                        regression_specs={'stars': None, 'dof': None, 'means': None, 'bracket_marker': 'angled'})

        table.print_table(fp='output/tables/tests_of_equality.tex')

    pass
