import pandas as pd
from pandas.tseries.offsets import Week


def check_primary_key_integrity(df, primary_key):
    """Check that df columns in primary_key consist no missing values or duplicates."""

    assert df[primary_key].notna().all().all(), 'Null values detected in primary key.'
    assert not df[primary_key].duplicated().any(), 'Duplicate values detected in primary key.'

    pass


def check_panel_integrity(df, freq, panel_vars):
    """
    Checks that time series in a panel is regularly distributed between its min and max dates,
    i.e. that no time series contain extra observations (e.g. monthly series has middle-of-month observation)
    or that there are gaps.

    Parameters
    ----------
    df: pandas.DataFrame
        A Pandas dataframe that includes date and panel variables as columns.
    freq: string
        Data frequency. Must be in ['day', 'week', 'month', 'quarter', 'year'].
        weekly series can be any weekday. Month and Quarter and must be end of period. Year must be end of any month.
        Alternative names: ['daily','weekly', 'monthly', 'quarterly', 'annual'].
    panel_vars: list
        First element specifies the name of date variable and second specifies the name of panel variable.

    Output
    ------
    Boolean

    Examples
    --------
    assert check_panel_integrity(df[['date', 'firm_id']], 'monthly', ['date', 'firm_id'])
    """
    assert isinstance(df, pd.DataFrame), 'df must be Pandas DataFrame object.'

    for i in df[panel_vars[1]].unique():
        temp_df = df[df[panel_vars[1]] == i]
        if check_ts_integrity(temp_df[panel_vars[0]], freq) is False:
            return False

    return True


def check_ts_integrity(series, freq):
    """
    Checks that a time series is regularly distributed between its min and max dates,
    i.e. that time series does not contain extra observations (e.g. monthly series has middle-of-month observation)
    or that there are gaps.

    Parameters
    ----------
    series: pandas.Series
        A datetime time series.
    freq: string
        Data frequency. Must be in ['day', 'week', 'month', 'quarter', 'year'].
        weekly series can be any weekday. Month and Quarter and must be end of period. Year must be end of any month.
        Alternative names: ['daily', 'weekly', 'monthly', 'quarterly', 'annual'].

    Output
    ------
    Boolean

    Examples
    --------
    assert check_ts_integrity(df['date'], 'monthly')
    """
    FREQS = {'day': 'D',
             'daily': 'D',
             'week': Week(1),
             'weekly': Week(1),
             'month': 'M',
             'monthly': 'M',
             'quarter': 'Q',
             'quarterly': 'Q',
             'year': 'A-{}',
             'annual': 'A-{}'
             }
    MONTHS = {1: 'JAN', 2: 'FEB', 3: 'MAR', 4: 'APR', 5: 'MAY', 6: 'JUN',
              7: 'JUL', 8: 'AUG', 9: 'SEP', 10: 'OCT', 11: 'NOV', 12: 'DEC'}

    assert isinstance(series, pd.Series), 'series must be Pandas Series object.'

    assert freq in FREQS.keys(), \
        "freq must be in {}.".format(str(list(FREQS.keys())))

    if freq == 'week' or freq == 'weekly':
        bmk_series = pd.date_range(series.min() - Week(1), series.max(), freq=FREQS[freq]).to_frame(index=False)
    elif freq == 'year' or freq == 'annual':
        bmk_series = pd.date_range(series.min(), series.max(),
                                   freq=FREQS[freq].format(MONTHS[series.min().month])).to_frame(index=False)
    else:
        bmk_series = pd.date_range(series.min(), series.max(), freq=FREQS[freq]).to_frame(index=False)

    bmk_series.rename(columns={0: series.name}, inplace=True)

    if bmk_series.shape[0] == series.shape[0]:
        if bmk_series.shape[0] == bmk_series.merge(series.to_frame(), on=series.name,
                                                   how='outer', indicator=True).shape[0]:
            return True

    return False
